/***BSD LICENSE***

Copyright (c) 2010, Zdenek Sikac
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
    * The name of the author may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

***END OF BSD LICENSE***


version 0.7

The program dsireg registers calibrated deep-sky images. It calculates transformation parameters of each image with respect to reference image, transforms images and saves them with prefix "reg~". Images can be filtered based on quality estimator - FWHM.

USE:

dsireg <switches> <input_file_1> <input_file_2> ... <input_file_N>

<input_file_1> - <input_file_N> images to be registered, image with best quality (lowest FWHM) is used as reference.

Switches:

search parameters:
-d [int]	minimum distance between stars (default 2) [pixels]
-sc [int] 	number of selected stars (default 200), minimum 25 is required
-tc [int] 	number of triangles (default 5000)
-td [float]	maximum sum of squared differences of sides between compared triangles (default 3) [pixels]

quality limits:
-q [float]	maximum FWHM of registered images (multiples of lowest FWHM) (default 1.5), minimum 1.1 is required

output log tables:
-it [string]	name of output file with table of images (default string = "dsireg_images_log")
-tt [string]	name of output file with time table (default no output)

*/

#include <iostream>
#include <fstream>
#include <Magick++.h>
#include <cmath>
#include <string>
#include <iomanip>
#include <algorithm>
#include <sstream>
#include "math.h"
#include <vector>

using namespace std;
using namespace Magick;

//default filenames parameters
const string def_time_table_name = "/dev/null";
const string def_images_table_name = "dsireg_images_log"; 

//default quality filter
const float def_max_fwhm = 1.5;

//default parameters for searching of stars
const int def_min_distance = 2;
const int def_target_stars_count = 200;

//default parameters for construction of triangles
const int def_target_triangles_count = 5000;

//default parameters for matching triangles
const float def_max_triangle_diff = 3; //optimized for DSLR Bayer matrix

//parameters for evaluation of regression
const int def_min_fits = 10;
const float def_max_res_sd = 1;

//comparator functions;
bool GR (float a,float b) {return (a>b);}

template <class T> bool GR_br (T a,T b) {return (a.get_br()>b.get_br());}
template <class T> bool GR_side_c (T a,T b) {return (a.get_side_c()>b.get_side_c());}

//Star: position x,y and brightness br
class Star {
float x,y,br;
float fwhm;
public:
void set (float,float,float,float);
void set (float,float,float);
void set (float,float);
float get_x();
float get_y();
float get_br();
float get_fwhm();
};//class Star 

void Star::set (float xi, float yi, float bri, float fwhmi) {x = xi; y = yi; br = bri; fwhm = fwhmi;}
void Star::set (float xi, float yi) {x = xi; y = yi;}
float Star::get_x() {return(x);}
float Star::get_y() {return(y);}
float Star::get_br() {return(br);}
float Star::get_fwhm() {return(fwhm);}

//Triangle defined by stars
class Triangle {
float side_a,side_b,side_c;
Star starA,starB,starC;
public:
void set (Star, Star, Star);
void arrange();
float get_side_a() {return(side_a);};
float get_side_b() {return(side_b);};
float get_side_c() {return(side_c);};
float get_A_x() {return(starA.get_x());};
float get_A_y() {return(starA.get_y());};
float get_B_x() {return(starB.get_x());};
float get_B_y() {return(starB.get_y());};
float get_C_x() {return(starC.get_x());};
float get_C_y() {return(starC.get_y());};
};//class Triangle

void Triangle::set (Star ai, Star bi, Star ci) {starA = ai;starB = bi;starC = ci;}

//set same orientation of triangles 
void Triangle::arrange() {

Vector <float> tmp_A(starA.get_x(),starA.get_y());
Vector <float> tmp_B(starB.get_x(),starB.get_y());
Vector <float> tmp_C(starC.get_x(),starC.get_y());

//calculate sides of triangle
Vector <float> tmp_AC = subtract(tmp_C, tmp_A);
Vector <float> tmp_AB = subtract(tmp_B, tmp_A);
Vector <float> tmp_BC = subtract(tmp_C, tmp_B);
float tmp_side_a = tmp_BC.size();
float tmp_side_b = tmp_AC.size();
float tmp_side_c = tmp_AB.size();

//assign A,B,C points so that sides a>=b>=c
if ((tmp_side_a>=tmp_side_b)&&(tmp_side_a>=tmp_side_c)) {
	starA.set(tmp_A.get_x1(),tmp_A.get_x2());
		if (tmp_side_b>=tmp_side_c) {starB.set(tmp_B.get_x1(),tmp_B.get_x2());starC.set(tmp_C.get_x1(),tmp_C.get_x2());}
			else {starB.set(tmp_C.get_x1(),tmp_C.get_x2());starC.set(tmp_B.get_x1(),tmp_B.get_x2());}
	}//if ((tmp_side_a>=tmp_side_b)&&(tmp_side_a>=tmp_side_c))
if ((tmp_side_b>=tmp_side_a)&&(tmp_side_b>=tmp_side_c)) {
	starA.set(tmp_B.get_x1(),tmp_B.get_x2());
		if (tmp_side_a>=tmp_side_c) {starB.set(tmp_A.get_x1(),tmp_A.get_x2());starC.set(tmp_C.get_x1(),tmp_C.get_x2());}
			else {starB.set(tmp_C.get_x1(),tmp_C.get_x2());starC.set(tmp_A.get_x1(),tmp_A.get_x2());}
	}//if ((tmp_side_b>=tmp_side_a)&&(tmp_side_b>=tmp_side_c))
if ((tmp_side_c>=tmp_side_a)&&(tmp_side_c>=tmp_side_b)) {
	starA.set(tmp_C.get_x1(),tmp_C.get_x2());
		if (tmp_side_a>=tmp_side_b) {starB.set(tmp_A.get_x1(),tmp_A.get_x2());starC.set(tmp_B.get_x1(),tmp_B.get_x2());}
			else {starB.set(tmp_B.get_x1(),tmp_B.get_x2());starC.set(tmp_A.get_x1(),tmp_A.get_x2());}
	}//if ((tmp_side_c>=tmp_side_a)&&(tmp_side_c>=tmp_side_b))

//assign values to Triangle class members - the stars 
Vector <float> pointA(starA.get_x(),starA.get_y());
Vector <float> pointB(starB.get_x(),starB.get_y());
Vector <float> pointC(starC.get_x(),starC.get_y());

Vector <float> AC = subtract(pointC, pointA);
Vector <float> AB = subtract(pointB, pointA);
Vector <float> BC = subtract(pointC, pointB);

side_a = BC.size();
side_b = AC.size();
side_c = AB.size();
}//void Triangle::arrange()

class LocalOffset {
float x,y,x2,y2;//star position and powers
float dx,dy;//star translations vs. reference image
float dx_res,dy_res;//residues from regression
bool valid;

public:
void set(float,float,float,float,bool);void set_valid(bool);void set_residues (float,float);
float get_x(),get_y(),get_x2(),get_y2(),get_dx(),get_dy(),get_dx_res(),get_dy_res();
bool get_valid();
};//class LocalOffset

void LocalOffset::set (float xi,float yi,float dxi,float dyi,bool validi) {
x=xi;y=yi;dx=dxi;dy=dyi;valid=validi;x2=square(xi);y2=square(yi);
};//void LocalOffset::set (float xi,float yi,float dxi,float dyi,bool validi)

void LocalOffset::set_valid (bool validi) {
valid=validi;
};//void LocalOffset::set_valid (bool validi)

void LocalOffset::set_residues (float dx_res_i,float dy_res_i) {
dx_res=dx_res_i;dy_res=dy_res_i;
};//void LocalOffset::set_residues (float dx_res_i,float dy_res_i)

float LocalOffset::get_x() {return(x);}
float LocalOffset::get_y() {return(y);}
float LocalOffset::get_x2() {return(x2);}
float LocalOffset::get_y2() {return(y2);}
float LocalOffset::get_dx() {return(dx);}
float LocalOffset::get_dy() {return(dy);}
float LocalOffset::get_dx_res() {return(dx_res);}
float LocalOffset::get_dy_res() {return(dy_res);}
bool LocalOffset::get_valid() {return(valid);}


class ImageAnalysis {
void initialize();
int width,height,area,depth,nil,im_median;
int *green;
float im_stdev,im_average,fwhm_median,br_loq;
int min_d,target_sc,target_tc,n_stars,n_triangles;
Image image;
vector <Star> stars;
vector <Triangle> triangles;

int pixel_position (int, int);

public:
ImageAnalysis (Image);
ImageAnalysis (string);
~ImageAnalysis();
void find_all_stars();
void select_valid_stars();
void find_triangles();
void save_stars(string);
void save_triangles(string);
void show_image_parameters();

void set_search_parameters(int min_d_i,int tsc_i,int ttc_i) {min_d=min_d_i;target_sc=tsc_i;target_tc=ttc_i;}
int get_width() {return(width);}
int get_height() {return(height);}
int get_area() {return(area);}
float get_average() {return(im_average);}
float get_stdev() {return(im_stdev);}
int get_median() {return(im_median);}
int get_n_stars() {return(n_stars);}
int get_n_triangles() {return(n_triangles);}
float get_br_loq() {return(br_loq);}
float get_fwhm_median() {return(fwhm_median);}
int get_min_d() {return(min_d);}
int get_target_sc() {return(target_sc);}
int get_target_tc() {return(target_tc);}
int get_star_br(int c) {return stars[c].get_br();}
float get_triangle_side_c(int c) {return triangles[c].get_side_c();}
Triangle get_triangle(int c) {return (triangles[c]);}
};

//position of pixel with coordinates x,y in memory
int ImageAnalysis::pixel_position (int x, int y) {
return (x + y*width);
}//int ImageAnalysis::pixel_position (int x, int y) 

ImageAnalysis::~ImageAnalysis () {
delete [] green;
}//ImageAnalysis::~ImageAnalysis ()

ImageAnalysis::ImageAnalysis (Image img) {
image = img;
initialize();
}//ImageAnalysis::ImageAnalysis (Image img)

ImageAnalysis::ImageAnalysis (string filename) {

cout<<"Opening file "<<filename<<"\n";

Image new_image;

try {
	new_image.read(filename);
	}

catch ( Exception &error_ ) {
	cout << "Caught exception: " << error_.what() << endl;
	
	}//try
image = new_image;
initialize();
}

//calculate parameters of image
void ImageAnalysis::initialize () {
width=image.columns();height=image.rows();
area=width*height; //area = number of pixels
const float base = 2;

depth=image.depth();nil=int(pow(base,depth));//nil=number of intensity levels

image.modifyImage();
Pixels pixelcache(image);
PixelPacket *pixels=pixelcache.get(0,0,width,height);//pixel cache from image

//get array of pixel intensities in sample from image
green = new int [area];
for (int c=0;c<area;c++) {
	green[c]=pixels[c].green;
	};//for (int c=0;c<area;c++)

pixelcache.sync();

im_average = average (green, area);
im_stdev = stdev (green, area);
im_median = quantil (green, area, 0.5);
br_loq=im_median+10*sqrt(im_median);
pixelcache.sync();
}//ImageAnalysis::ImageAnalysis (string filename)

void ImageAnalysis::find_all_stars () {

stars.clear();

int br_0, br_comp; //intensity values of tested pixel and compared pixel
bool is_star;
int br_sum;
float *br_sur;
float *d_br;
float *d2_br;
float fwhm,x_sub,y_sub;
int r_tmp;
Star tmp_star;

br_sur = new float [min_d];
d_br = new float [min_d];
d2_br = new float [min_d];

cout<<"Counting stars";

for (int x=min_d; x<(width-min_d);x++) {
	for (int y=min_d; y<(height-min_d);y++) {
		is_star=true;
		br_0 = green[pixel_position(x,y)];//evaluated pixel
		r_tmp = 1;
		if (br_0 < br_loq) continue; //exclude low intensity pixels
		for (int dx=-min_d; dx<=min_d; dx++) {
			for (int dy=-min_d; dy<=min_d; dy++) {
				if ((dx==0)&&(dy==0)) continue;
				br_comp = green[pixel_position(x+dx,y+dy)];
				if (br_comp>=br_0) {is_star=false;break;}; //not a star if brighter point is near
				} //for (int dy=-min_d; dy<=min_d; dy++)
			if (is_star==false) break;
			} //for (int dx=-min_d; dx<=min_d; dx++)
		
		if (is_star==false) continue; //exit search if no star is found

		//test for star shape
		for (int r=0; r<min_d; r++) {	//measure surrounding intensity
			br_sum = green[pixel_position(x-r,y)];
			br_sum = br_sum + green[pixel_position(x+r,y)];
			br_sum = br_sum + green[pixel_position(x,y-r)];
			br_sum = br_sum + green[pixel_position(x,y+r)];
			br_sur[r] = br_sum/4;
			} //for (int r=0; r<min_d; r++)

		for (int r=0; r<min_d; r++) {	//calculate gradient, first derivative
			d_br[r] = br_sur[r+1]-br_sur[r];
			} //for (int r=0; r<min_d; r++)

		for (int r=0; r<min_d; r++) {	//second derivative
			d2_br[r] = d_br[r+1]-d_br[r];
			} //for (int r=0; r<min_d; r++)

		for (int r=1; r<min_d; r++) {	//detect inflection
			if(d2_br[r]>0) {r_tmp = r;break;}
			} //for (int r=1; r<min_d; r++)
		
		if (d_br[0]<(3*im_median-br_sur[0])) {is_star = false;continue;} //detect hotpixels
		if (br_sur[1]>(br_sur[0]*0.9)) {is_star = false;continue;} //detect overburnt stars
		if (is_star==false) continue; //exit search if no star is found

		fwhm=r_tmp*br_sur[0]/(br_sur[0]-br_sur[r_tmp]);//estimate FWHM from tangent at inflection
		x_sub=0.5*(green[pixel_position(x+1,y)]-green[pixel_position(x-1,y)])/green[pixel_position(x,y)];
		y_sub=0.5*(green[pixel_position(x,y+1)]-green[pixel_position(x,y-1)])/green[pixel_position(x,y)];
		tmp_star.set(x+x_sub,y+y_sub,br_sur[0],fwhm);
		stars.push_back(tmp_star);
		} //for (int y=min_d; y<(height-min_d);y++) 
	counter (x,width-min_d,10);
	} //for (int x=min_d; x<(width-min_d);x++)

n_stars=stars.size();

delete [] br_sur;
delete [] d_br;
delete [] d2_br;

cout<<n_stars<<" stars have been found.\n";

if (n_stars==0) {return;}// exit if no stars are found

float *fwhm_tmp;
fwhm_tmp = new float[n_stars];
for (int c=0;c<n_stars;c++) {
	fwhm_tmp[c] = stars[c].get_fwhm();
	}//for (int c=0;c<n_stars;c++)
fwhm_median = quantil (fwhm_tmp,n_stars,0.5);
delete[] fwhm_tmp;
}//void ImageAnalysis::find_all_stars ()

void ImageAnalysis::select_valid_stars() {
sort(stars.begin(),stars.end(),GR_br <Star>);
stars.erase(stars.begin()+target_sc,stars.end());
n_stars=stars.size();
cout<<"Limited to "<<n_stars<<" stars brighter than "<<stars.end()->get_br()<<".\n";
}//void ImageAnalysis::select_valid_stars()

//create triangles from identified stars
void ImageAnalysis::find_triangles () {

triangles.clear();
float tmp_side;
Triangle tmp_triangle;

if (n_stars<3) {n_triangles=0;return;} //exit if no triangles can be constructed

cout<<"Constructing triangles";
for (int c1=0;c1<(n_stars-2);c1++) {
	counter (c1,n_stars-2,10);
	for (int c2=c1+1;c2<(n_stars-1);c2++) {
		for (int c3=c2+1;c3<n_stars;c3++) {
			tmp_triangle.set (stars[c1],stars[c2],stars[c3]);
			tmp_triangle.arrange();
			tmp_side=tmp_triangle.get_side_c();
			triangles.push_back(tmp_triangle);
			}//for (int c3=c2+1;c3<n_stars;c3++)
		}//for (int c2=c1+1;c2<(n_stars-1);c2++)
	}//for (int c1=0;c1<(n_stars-2);c1++)
n_triangles=triangles.size();
cout<<n_triangles<<" triangles have been found.\n";

sort(triangles.begin(),triangles.end(),GR_side_c <Triangle>);
triangles.erase(triangles.begin()+target_tc,triangles.end());
n_triangles=triangles.size();

cout<<"Limited to "<<n_triangles<<" triangles with sides > "<<triangles.end()->get_side_c()<<" pixels.\n";
}//void ImageAnalysis::find_triangles ()

//save stars into text file
void ImageAnalysis::save_stars(string filename) {
ofstream myfile;
const char *cfilename = filename.c_str();

myfile.open (cfilename);
cout<<"Saving stars to file "<<cfilename;
myfile<<setw(8)<<"N:"<<setw(10)<<"x:"<<setw(10)<<"y:"<<setw(10)<<"BR:"<<setw(10)<<"FWHM:\n";
for (int c=0; c<n_stars;c++) {
	myfile<<setw(8)<<fixed<<showpoint<<setprecision(0)<<c;
	myfile<<setw(10)<<setprecision(2)<<stars[c].get_x();
	myfile<<setw(10)<<stars[c].get_y();
	myfile<<setw(10)<<setprecision(0)<<noshowpoint<<stars[c].get_br();
	myfile<<setw(10)<<setprecision(2)<<showpoint<<stars[c].get_fwhm()<<"\n";
	counter (c,n_stars,10);
	}//for (int c=0; c<n_stars;c++)
myfile.close();
cout<<"Saved.\n";
}//void ImageAnalysis::save_stars(string filename)

//save triangles into text file
void ImageAnalysis::save_triangles(string filename) {
ofstream myfile;
const char *cfilename = filename.c_str();

myfile.open (cfilename);
cout<<"Saving triangles to file";
myfile <<setw(8)<<"Number:"<<setw(10)<<"a:"<<setw(10)<<"b:"<<setw(10)<<"c:\n";
for (int c=0; c<n_triangles;c++) {
	myfile<<setw(8)<<fixed<<showpoint<<setprecision(0)<<c;
	myfile<<setw(10)<<setprecision(2)<<triangles[c].get_side_a();
	myfile<<setw(10)<<triangles[c].get_side_b();
	myfile<<setw(10)<<triangles[c].get_side_c()<<"\n";
	counter (c,n_triangles,10);
	}//for (int c=0; c<n_triangles;c++)
myfile.close();
cout<<"Done.\n";
}//void ImageAnalysis::save_triangles(string filename)

//display image parameters
void ImageAnalysis::show_image_parameters() {

cout <<"width = "<<get_width()<<"\n";
cout <<"height = "<<get_height()<<"\n";
cout <<"area = "<<get_area()<<"\n";
cout <<"average = "<<get_average()<<"\n";
cout <<"stdev = "<<get_stdev()<<"\n";
cout <<"median = "<<get_median()<<"\n";
cout <<"FWMH = "<<get_fwhm_median()<<"\n";
}//void ImageAnalysis::show_image_parameters()

//compare image to reference image and calculate offset and rotation
class FitImages {
ImageAnalysis *ref,*com;
float cdx_x2,cdx_y2,cdx_x,cdx_y,cdx_1;//regression coefficients for dx
float cdy_x2,cdy_y2,cdy_x,cdy_y,cdy_1;//regression coefficients for dy
float dx_res_avg,dy_res_avg,dx_res_sd,dy_res_sd;//averages and standard deviations of dx and dy residues
int np;//total number of matched points
int npv;//number of valid matched points
int n_outliers;
LocalOffset *lo;
LocalOffset *lov;

public:
FitImages (ImageAnalysis&, ImageAnalysis&, float);
int get_np() {return(np);};
int get_npv() {return(npv);};
float get_cdx_x2() {return(cdx_x2);};
float get_cdx_y2() {return(cdx_y2);};
float get_cdx_x() {return(cdx_x);};
float get_cdx_y() {return(cdx_y);};
float get_cdx_1() {return(cdx_1);};
float get_cdy_x2() {return(cdy_x2);};
float get_cdy_y2() {return(cdy_y2);};
float get_cdy_x() {return(cdy_x);};
float get_cdy_y() {return(cdy_y);};
float get_cdy_1() {return(cdy_1);};
float get_dx_res_sd() {return(dx_res_sd);};
float get_dy_res_sd() {return(dy_res_sd);};
void save(string);
void extract_valid_points();
void remove_redundant_points();
void calculate_regression();
void invalidate_outliers();
};//class FitImages

FitImages::FitImages (ImageAnalysis& ref,ImageAnalysis& com, float max_triangle_diff_i) {
float x_tmp,y_tmp,dx_tmp,dy_tmp;
int hit=0;
float diff_tmp, diff_cur;
n_outliers=1;
lo = new LocalOffset[com.get_n_triangles()*3];//all points of triangles will be used

cout<<"Comparing triangles";
for (int c1=0;c1<com.get_n_triangles();c1++) {
	counter(c1,com.get_n_triangles(),10);
	int best_fit=-1;diff_tmp=max_triangle_diff_i;
	for (int c2=0;c2<ref.get_n_triangles();c2++) {
		diff_cur=square(com.get_triangle(c1).get_side_a()-ref.get_triangle(c2).get_side_a());		
		diff_cur=diff_cur+square(com.get_triangle(c1).get_side_b()-ref.get_triangle(c2).get_side_b());
		diff_cur=diff_cur+square(com.get_triangle(c1).get_side_c()-ref.get_triangle(c2).get_side_c());
		if (diff_cur<diff_tmp) {best_fit=c2;diff_tmp=diff_cur;}
		}//for (int c2=0;c2<ref.get_n_triangles();c2++)
	if (best_fit>-1) {
		x_tmp = com.get_triangle(c1).get_A_x();
		y_tmp = com.get_triangle(c1).get_A_y();
		dx_tmp = com.get_triangle(c1).get_A_x()-ref.get_triangle(best_fit).get_A_x();
		dy_tmp = com.get_triangle(c1).get_A_y()-ref.get_triangle(best_fit).get_A_y();
		lo[hit].set(x_tmp,y_tmp,dx_tmp,dy_tmp,true);
		hit++;

		x_tmp = com.get_triangle(c1).get_B_x();
		y_tmp = com.get_triangle(c1).get_B_y();
		dx_tmp = com.get_triangle(c1).get_B_x()-ref.get_triangle(best_fit).get_B_x();
		dy_tmp = com.get_triangle(c1).get_B_y()-ref.get_triangle(best_fit).get_B_y();
		lo[hit].set(x_tmp,y_tmp,dx_tmp,dy_tmp,true);
		hit++;

		x_tmp = com.get_triangle(c1).get_C_x();
		y_tmp = com.get_triangle(c1).get_C_y();
		dx_tmp = com.get_triangle(c1).get_C_x()-ref.get_triangle(best_fit).get_C_x();
		dy_tmp = com.get_triangle(c1).get_C_y()-ref.get_triangle(best_fit).get_C_y();
		lo[hit].set(x_tmp,y_tmp,dx_tmp,dy_tmp,true);
		hit++;
		}//if (best_fit>-1) 
	}//for (int c1=0;c1<com.get_n_triangles();c1++)
np = hit;


if(np==0) {cout<<"No matching triangles found.\n";
	cdx_x2=0;cdx_y2=0;cdx_x=0;cdx_y=0;cdx_1=0;
	cdy_x2=0;cdy_y2=0;cdy_x=0;cdy_y=0;cdy_1=0;
	dx_res_avg=0;dy_res_avg=0;dx_res_sd=0;dy_res_sd=0;
	npv=0;
	return;}//if(np==0)

LocalOffset *lo_aux;
lo_aux = new LocalOffset [np];
for (int c=0;c<np;c++) {
	lo_aux[c]=lo[c];	
	}//for (int c=0;c<np;c++)

delete[] lo;
lo = new LocalOffset [np];
for (int c=0;c<np;c++) {
	lo[c]=lo_aux[c];	
	}//for (int c=0;c<np;c++)
delete[] lo_aux;

cout<<"Total number of points: "<<np<<"\n";
remove_redundant_points();
cout<<"Number of unique points: "<<np<<"\n";

do {
	calculate_regression();
	invalidate_outliers();
	}//do
while (n_outliers>0);

cout<<"x-offset = "<<cdx_1<<" pixels, y-offset = "<<cdy_1<<" pixels.\n";
}//FitImages::FitImages (ImageAnalysis& ref,ImageAnalysis& com, float max_triangle_diff_i)

void FitImages::extract_valid_points() {
LocalOffset *lo_aux;
lo_aux = new LocalOffset [np];
int n=0;
for (int c=0;c<np;c++) {
	if(lo[c].get_valid()) {lo_aux[n]=lo[c];n++;}	
	}//for (int c=0;c<np;c++)
npv=n;

lov = new LocalOffset [npv];
for (int c=0;c<npv;c++) {
	lov[c]=lo_aux[c];	
	}//for (int c=0;c<npv;c++)
delete[] lo_aux;
}//void FitImages::extract_valid_points()

void FitImages::remove_redundant_points() {
float x_tmp,y_tmp;
for (int c1=0;c1<(np-1);c1++) {
	if(lo[c1].get_valid()==false) continue;	
	x_tmp=lo[c1].get_x();y_tmp=lo[c1].get_y();
	for (int c2=c1+1;c2<np;c2++) {
		if ((x_tmp==lo[c2].get_x())&&(y_tmp==lo[c2].get_y())) lo[c2].set_valid(false);
		}//for (int c2=c1+1;c2<np;c2++)
	}//for (int c1=0;c1<(np-1);c1++)

LocalOffset *lo_aux;
int n=0;
lo_aux = new LocalOffset [np];
for (int c=0;c<np;c++) {
	if(lo[c].get_valid()) {
		lo_aux[n]=lo[c];n++;
		}//if(lo[c].get_valid())
	}//for (int c=0;c<np;c++)
np=n;

delete[] lo;
lo=new LocalOffset[np];
for (int c=0;c<np;c++) {
	lo[c]=lo_aux[c];
	}//for (int c=0;c<np;c++)
delete[] lo_aux;
cout<<"Redundant points removed.\n";
}//void FitImages::remove_redundant_points()

void FitImages::calculate_regression() {
extract_valid_points();
cout<<"Number of valid points: "<<npv<<"\n";

float *dx,*dy,*x,*y,*x2,*y2,*one;
dx = new float[npv];
dy = new float[npv];
x = new float[npv];
y = new float[npv];
x2 = new float[npv];
y2 = new float[npv];
one = new float[npv];

for (int c=0;c<npv;c++) {
	dx[c] = lov[c].get_dx();
	dy[c] = lov[c].get_dy();
	x[c] = lov[c].get_x();
	y[c] = lov[c].get_y();
	x2[c] = lov[c].get_x2();
	y2[c] = lov[c].get_y2();
	one[c] = 1;
	}//for (int c=0;c<npv;c++)

double a_dx[] = {
scp(x2,x2,npv),scp(y2,x2,npv),scp(x,x2,npv),scp(y,x2,npv),scp(one,x2,npv),scp(dx,x2,npv),
scp(x2,y2,npv),scp(y2,y2,npv),scp(x,y2,npv),scp(y,y2,npv),scp(one,y2,npv),scp(dx,y2,npv),
scp(x2,x,npv),scp(y2,x,npv),scp(x,x,npv),scp(y,x,npv),scp(one,x,npv),scp(dx,x,npv),
scp(x2,y,npv),scp(y2,y,npv),scp(x,y,npv),scp(y,y,npv),scp(one,y,npv),scp(dx,y,npv),
scp(x2,one,npv),scp(y2,one,npv),scp(x,one,npv),scp(y,one,npv),scp(one,one,npv),scp(dx,one,npv)
};

Matrix Reg_dx(5,6,a_dx);
Reg_dx.separate();

cdx_x2=Reg_dx.get_a(1,6);
cdx_y2=Reg_dx.get_a(2,6);
cdx_x=Reg_dx.get_a(3,6);
cdx_y=Reg_dx.get_a(4,6);
cdx_1=Reg_dx.get_a(5,6);

cout<<"Coefficients of dx:\n";
cout<<"c_x2 = "<<cdx_x2<<"\n";
cout<<"c_y2 = "<<cdx_y2<<"\n";
cout<<"c_x = "<<cdx_x<<"\n";
cout<<"c_y = "<<cdx_y<<"\n";
cout<<"c_1 = "<<cdx_1<<"\n\n";

double a_dy[] = {
scp(x2,x2,npv),scp(y2,x2,npv),scp(x,x2,npv),scp(y,x2,npv),scp(one,x2,npv),scp(dy,x2,npv),
scp(x2,y2,npv),scp(y2,y2,npv),scp(x,y2,npv),scp(y,y2,npv),scp(one,y2,npv),scp(dy,y2,npv),
scp(x2,x,npv),scp(y2,x,npv),scp(x,x,npv),scp(y,x,npv),scp(one,x,npv),scp(dy,x,npv),
scp(x2,y,npv),scp(y2,y,npv),scp(x,y,npv),scp(y,y,npv),scp(one,y,npv),scp(dy,y,npv),
scp(x2,one,npv),scp(y2,one,npv),scp(x,one,npv),scp(y,one,npv),scp(one,one,npv),scp(dy,one,npv)
};

Matrix Reg_dy(5,6,a_dy);
Reg_dy.separate();

cdy_x2=Reg_dy.get_a(1,6);
cdy_y2=Reg_dy.get_a(2,6);
cdy_x=Reg_dy.get_a(3,6);
cdy_y=Reg_dy.get_a(4,6);
cdy_1=Reg_dy.get_a(5,6);

cout<<"Coefficients of dy:\n";
cout<<"c_x2 = "<<cdy_x2<<"\n";
cout<<"c_y2 = "<<cdy_y2<<"\n";
cout<<"c_x = "<<cdy_x<<"\n";
cout<<"c_y = "<<cdy_y<<"\n";
cout<<"c_1 = "<<cdy_1<<"\n";

//calculate residues
float x2_tmp,y2_tmp,x_tmp,y_tmp,dx_tmp,dy_tmp,dx_res_tmp,dy_res_tmp;
for (int c=0;c<np;c++) {
	dx_tmp = lo[c].get_dx();
	dy_tmp = lo[c].get_dy();
	x_tmp = lo[c].get_x();
	y_tmp = lo[c].get_y();
	x2_tmp = lo[c].get_x2();
	y2_tmp = lo[c].get_y2();
	dx_res_tmp = dx_tmp-(cdx_x2*x2_tmp+cdx_y2*y2_tmp+cdx_x*x_tmp+cdx_y*y_tmp+cdx_1);
	dy_res_tmp = dy_tmp-(cdy_x2*x2_tmp+cdy_y2*y2_tmp+cdy_x*x_tmp+cdy_y*y_tmp+cdy_1);
	lo[c].set_residues(dx_res_tmp,dy_res_tmp);	
	}//for (int c=0;c<np;c++)
}//void FitImages::calculate_regression()

void FitImages::invalidate_outliers() {
extract_valid_points();

int n=0;
float *dx_res_tmp,*dy_res_tmp;
dx_res_tmp = new float[npv];
dy_res_tmp = new float[npv];

for (int c=0;c<npv;c++) {
	dx_res_tmp[c]=lov[c].get_dx_res();
	dy_res_tmp[c]=lov[c].get_dy_res();
	}//for (int c=0;c<npv;c++)

dx_res_avg = average(dx_res_tmp,npv);
dx_res_sd = stdev(dx_res_tmp,npv);
dy_res_avg = average(dy_res_tmp,npv);
dy_res_sd = stdev(dy_res_tmp,npv);

delete[]dx_res_tmp;
delete[]dy_res_tmp;

cout<<"Residues of dx: AVG = "<<dx_res_avg<<"\tSD = "<<dx_res_sd<<"\n";
cout<<"Residues of dy: AVG = "<<dy_res_avg<<"\tSD = "<<dy_res_sd<<"\n";

for (int c=0;c<np;c++) {
	if (lo[c].get_valid()==false) continue;
	if (abs(lo[c].get_dx_res()-dx_res_avg)>(3*dx_res_sd)) {lo[c].set_valid(false);n++;continue;};
	if (abs(lo[c].get_dy_res()-dy_res_avg)>(3*dy_res_sd)) {lo[c].set_valid(false);n++;continue;};
	}//for (int c=0;c<np;c++)

cout<<n<<" outliers removed.\n\n";
n_outliers=n;
}//void FitImages::invalidate_outliers()

void FitImages::save(string filename) {
ofstream myfile;
const char *cfilename = filename.c_str();

myfile.open (cfilename);
cout<<"Saving local offsets to file "<<filename;
myfile<<setw(5)<<"N:"<<setw(10)<<"x:"<<setw(10)<<"y:"<<setw(10)<<"dx:"<<setw(10)<<"dy:";
myfile<<setw(10)<<"dx_res:"<<setw(10)<<"dy_res:"<<setw(10)<<"valid:\n";
for (int c=0; c<np;c++) {
	myfile<<noshowpos<<noshowpoint<<fixed<<setprecision(0)<<setw(5)<<c;
	myfile<<setw(10)<<setprecision(2)<<showpoint<<lo[c].get_x();
	myfile<<setw(10)<<lo[c].get_y();
	myfile<<setw(10)<<showpos<<lo[c].get_dx();
	myfile<<setw(10)<<lo[c].get_dy();
	myfile<<setw(10)<<scientific<<lo[c].get_dx_res();
	myfile<<setw(10)<<lo[c].get_dy_res();
	myfile<<setw(10)<<noshowpos<<setprecision(0)<<lo[c].get_valid()<<"\n";
	counter (c,np,10);
	}//for (int c=0; c<np;c++)
myfile.close();
cout<<"Saved.\n";
}//void FitImages::save(string filename)

//class for storing of image data
class FileRecord {
string filename;
float fwhm;
bool use, ref;
float cdx_x2,cdx_y2,cdx_x,cdx_y,cdx_1,dx_res_sd;
float cdy_x2,cdy_y2,cdy_x,cdy_y,cdy_1,dy_res_sd;
int n_stars,n_triangles,n_fits;

public:
void initialize (string);
void analyze_image(int,int);
void calculate_fwhm(int);
string get_filename() {return(filename);}
float get_fwhm() {return(fwhm);}
float get_cdx_x2() {return(cdx_x2);}
float get_cdx_y2() {return(cdx_y2);}
float get_cdx_x() {return(cdx_x);}
float get_cdx_y() {return(cdx_y);}
float get_cdx_1() {return(cdx_1);}
float get_dx_res_sd() {return(dx_res_sd);}
float get_cdy_x2() {return(cdy_x2);}
float get_cdy_y2() {return(cdy_y2);}
float get_cdy_x() {return(cdy_x);}
float get_cdy_y() {return(cdy_y);}
float get_cdy_1() {return(cdy_1);}
float get_dy_res_sd() {return(dy_res_sd);}
int get_n_fits() {return(n_fits);}
int get_n_stars() {return(n_stars);}
int get_n_triangles() {return(n_triangles);}
bool get_use() {return(use);}
bool get_ref() {return(ref);}
void set_fwhm(float fwhm_i) {fwhm=fwhm_i;}
void set_ref(bool new_ref) {ref = new_ref;}
void set_n_stars(int ns_i) {n_stars=ns_i;}
void set_n_triangles(int nt_i) {n_triangles=nt_i;}
void set_use(bool new_use) {use = new_use;}
void set_trafo_parameters(float,float,float,float,float,float,float,float,float,float,float,float,int);
void transform_image();
FileRecord();
};//class FileRecord

FileRecord::FileRecord() {
filename="";fwhm=0;use=true;ref=false;
cdx_x2=0;cdx_y2=0;cdx_x=0;cdx_y=0;cdx_1=0;
cdy_x2=0;cdy_y2=0;cdy_x=0;cdy_y=0;cdy_1=0;
n_stars=0;n_triangles=0;n_fits=0;
}//FileRecord::FileRecord()

void FileRecord::initialize(string filename_i) {
filename=filename_i;
fwhm=0;use=true;ref=false;
cdx_x2=0;cdx_y2=0;cdx_x=0;cdx_y=0;cdx_1=0;
cdy_x2=0;cdy_y2=0;cdy_x=0;cdy_y=0;cdy_1=0;
n_stars=0;n_triangles=0;n_fits=0;
}//void FileRecord::intialize(string filename_i)

void FileRecord::calculate_fwhm(int min_d) {
ImageAnalysis test_ia(filename);
test_ia.set_search_parameters(min_d,0,0);
test_ia.find_all_stars();
fwhm=test_ia.get_fwhm_median();
n_stars=test_ia.get_n_stars();
}//void FileRecord::calculate_fwhm(int min_d)

void FileRecord::set_trafo_parameters
(float cdx_x2_i,float cdx_y2_i,float cdx_x_i,float cdx_y_i,float cdx_1_i,float dx_res_sd_i,
float cdy_x2_i,float cdy_y2_i,float cdy_x_i,float cdy_y_i,float cdy_1_i,float dy_res_sd_i,
int n_fits_i) {

cdx_x2=cdx_x2_i;cdx_y2=cdx_y2_i;cdx_x=cdx_x_i;cdx_y=cdx_y_i;cdx_1=cdx_1_i;dx_res_sd=dx_res_sd_i;
cdy_x2=cdy_x2_i;cdy_y2=cdy_y2_i;cdy_x=cdy_x_i;cdy_y=cdy_y_i;cdy_1=cdy_1_i;dy_res_sd=dy_res_sd_i;

n_fits=n_fits_i;
}//void FileRecord::set_trafo_parameters

void FileRecord::transform_image() {
Image original_image (filename);
int w = original_image.columns();int h = original_image.rows();
int area=w*h;

Image transformed_image(Geometry(w,h),Color(0,0,0,MaxRGB));

float x_tmp,y_tmp;

original_image.modifyImage();
Pixels original_cache(original_image);
PixelPacket *original_pixels=original_cache.get(0,0,w,h);//pixel cache from original image

transformed_image.modifyImage();
Pixels transformed_cache(transformed_image);
PixelPacket *transformed_pixels=transformed_cache.get(0,0,w,h);//pixel cache from image

int cx,cy;
int pos_tmp;

cout<<"Transforming image "<<filename;

for (int c=0;c<area;c++) {
	counter (c,area,10);	
	cx=c%w;	
	cy=c/w;
	x_tmp = cx+(cdx_x2*square(cx)+cdx_y2*square(cy)+cdx_x*cx+cdx_y*cy+cdx_1);
	y_tmp = cy+(cdy_x2*square(cx)+cdy_y2*square(cy)+cdy_x*cx+cdy_y*cy+cdy_1);
	pos_tmp = round(x_tmp)+round(y_tmp)*w;
	if ((pos_tmp<0)||(pos_tmp>area)) continue;
	transformed_pixels[c] = original_pixels[pos_tmp];
	}//for (int c=0;c<area;c++)

original_cache.sync();
transformed_cache.sync();

transformed_image.matte(false);
transformed_image.write("reg_"+filename);
cout<<"Saved as reg_"<<filename<<".\n";
}//void FileRecord::transform_image()

//table for storing of FileRecords
class FileTable {
int n_files;FileRecord *frecord;int min_distance;float triangle_max_diff; int ref_index;
int target_sc,target_tc;

public:
int get_n_files() {return (n_files);};
void find_ref();
void calculate_transformations();
void analyze_images();
void calculate_fwhm();
FileTable(int,int,int,int,float);
~FileTable();
void set_frecord(int, string);
void save(string);
int get_ref_index() {return(ref_index);}
void deselect_nonfits(int);
void deselect_bad_fits(float);
void deselect_high_fwhm(float);
void deselect_low_stars(int);
FileRecord get_frecord(int c) {return frecord[c];}
void transform_images();
};//class FileTable;

FileTable::FileTable (int n_files_i,int min_distance_i,int target_sc_i,int target_tc_i,float triangle_max_diff_i) {
n_files = n_files_i;ref_index = 0;
frecord = new FileRecord [n_files];
min_distance = min_distance_i;
target_sc=target_sc_i;
target_tc=target_tc_i;
triangle_max_diff = triangle_max_diff_i;
}//FileTable::FileTable (int n_files_i,int min_distance_i,float star_threshold_i,float triangle_min_side_i,float triangle_max_diff_i)

FileTable::~FileTable() {
delete [] frecord;
}//FileTable::~FileTable()

//assign filenames to file records
void FileTable::set_frecord (int n, string filename) {
frecord[n].initialize (filename);
}//void FileTable::set_frecord (int n, string filename)

//calculate FWHM of all images with stars s/n > 10 and set search criteria for stars and triangles 
void FileTable::calculate_fwhm() {
cout<<"Measuring FWHM. Search limited to stars with s/n > 10.\n";
for (int c=0;c<n_files;c++) {
	frecord[c].calculate_fwhm(min_distance);
	cout<<"Median FWHM: "<<frecord[c].get_fwhm()<<" pixels.\n";
	}//for (int c=0;c<n_files;c++)
}//void FileTable::analyze_images()

//find reference image (with lowest FWHM)
void FileTable::find_ref() { 
float min_fwhm = frecord[0].get_fwhm();
float tmp_fwhm = 0;
for (int c=0; c<n_files;c++) {
	frecord[c].set_ref(false);
	tmp_fwhm = frecord[c].get_fwhm();
	if(tmp_fwhm<min_fwhm) {ref_index = c;min_fwhm = tmp_fwhm;}
	}//for (int c=0; c<n_files;c++)
frecord[ref_index].set_ref(true);
}//void FileTable::find_ref() 

//calculate transformation parameters from reference image
void FileTable::calculate_transformations() {
string ref_file, com_file;
ref_file = frecord[ref_index].get_filename();
cout<<"Using image "<<ref_file<<" as reference.\n";
ImageAnalysis ref_ia(ref_file);
ref_ia.set_search_parameters(min_distance,target_sc,target_tc);
ref_ia.find_all_stars();
ref_ia.select_valid_stars();
ref_ia.find_triangles();
ref_ia.save_stars(ref_file+".stars");
 
for (int c=0; c<n_files;c++) {
	if(c==ref_index) continue; //skip calculation if current image is reference
	if(frecord[c].get_use()==false) continue; //skip calculation if current image is marked as unused
	com_file=frecord[c].get_filename();
	cout<<"Calculating offsets of image "<<com_file<<" from reference image.\n";
	ImageAnalysis com_ia(com_file);
	com_ia.set_search_parameters(min_distance,target_sc,target_tc);
	com_ia.find_all_stars();
	com_ia.select_valid_stars();
	com_ia.find_triangles();
	com_ia.save_stars(com_file+".stars");

	FitImages fit(ref_ia,com_ia,triangle_max_diff);
	frecord[c].set_trafo_parameters
		(fit.get_cdx_x2(),fit.get_cdx_y2(),fit.get_cdx_x(),fit.get_cdx_y(),fit.get_cdx_1(),fit.get_dx_res_sd(),
		 fit.get_cdy_x2(),fit.get_cdy_y2(),fit.get_cdy_x(),fit.get_cdy_y(),fit.get_cdy_1(),fit.get_dy_res_sd(),
		 fit.get_npv());
	fit.save(com_file+".offsets");
	}//for (int c=0; c<n_files;c++)
}//void FileTable::calculate_transformations()

//deselect images with low count of fitting triangles
void FileTable::deselect_nonfits(int lim) {
for (int c=0; c<n_files;c++) {
	int nf=frecord[c].get_n_fits();
	bool r=frecord[c].get_ref();
	if(nf<lim&&r==false) frecord[c].set_use(false);
	}//for (int c=0; c<n_files;c++) 
}//void FileTable::deselect_nonfits(lim)

//deselect images with high FWHM
void FileTable::deselect_high_fwhm(float lim) {
cout<<"Deselecting images with high FWHM.\n";
float ref_fwhm = frecord[ref_index].get_fwhm(); 
for (int c=0; c<n_files;c++) {
	float cur_fwhm = frecord[c].get_fwhm();
	if((cur_fwhm/ref_fwhm)>lim) frecord[c].set_use(false);
	}//for (int c=0; c<n_files;c++)
}//void FileTable::deselect_high_fwhm(float lim)

//deselect images with low number of stars
void FileTable::deselect_low_stars(int lim) {
cout<<"Deselecting images with low stars count.\n";
for (int c=0; c<n_files;c++) {
	float ns = frecord[c].get_n_stars();
	if(ns<lim) frecord[c].set_use(false);
	}//for (int c=0; c<n_files;c++)
}//void FileTable::deselect_low_stars(int lim)

//deselect bad fits
void FileTable::deselect_bad_fits(float lim) {
cout<<"Deselecting images with high SD of residues.\n";
for (int c=0; c<n_files;c++) {
	float dxrsd = frecord[c].get_dx_res_sd();
	float dyrsd = frecord[c].get_dy_res_sd();
	if(dxrsd>lim) frecord[c].set_use(false);
	if(dyrsd>lim) frecord[c].set_use(false);
	}//for (int c=0; c<n_files;c++)
}//void FileTable::deselect_bad_fits(float lim)

//register images
void FileTable::transform_images() {
for (int c=0; c<n_files;c++) {
	if(frecord[c].get_use()==false) continue;
	frecord[c].transform_image();
	}//for (int c=0; c<n_files;c++)
}//void FileTable::transform_images()

//save file table into file
void FileTable::save(string file) {
ofstream myfile;
const char *cfilename = file.c_str();

myfile.open (cfilename);

cout<<"Saving table of images to file "<<file;
myfile<<setw(5)<<"N:"<<setw(20)<<"File:"<<setw(8)<<"Ref:"<<setw(8)<<"Used:"<<setw(8);
myfile<<setw(8)<<"FWHM:"<<setw(8)<<"Stars:"<<setw(8)<<"Fits:";
myfile<<setw(12)<<"dx_res_sd:"<<setw(12)<<"dy_res_sd:"<<setw(12)<<"cdx_1:"<<setw(12)<<"cdy_1:"<<setw(12)<<"cdx_x:"<<setw(12)<<"cdy_x:";
myfile<<setw(12)<<"cdx_y:"<<setw(12)<<"cdy_y:";
myfile<<setw(12)<<"cdx_x2:"<<setw(12)<<"cdy_x2:"<<setw(12)<<"cdx_y2:"<<setw(12)<<"cdy_y2:\n";

for (int c=0; c<n_files;c++) {
	myfile<<setw(5)<<fixed<<setprecision(0)<<noshowpos<<c;
	myfile<<setw(20)<<frecord[c].get_filename();
	myfile<<setw(8)<<setprecision(0)<<frecord[c].get_ref()<<setw(8)<<frecord[c].get_use();
	myfile<<setw(8)<<showpoint<<setprecision(2)<<setw(8)<<frecord[c].get_fwhm();
	myfile<<setw(8)<<frecord[c].get_n_stars();
	myfile<<setw(8)<<setprecision(0)<<frecord[c].get_n_fits();
	myfile<<setw(12)<<setprecision(2)<<scientific<<showpos<<frecord[c].get_dx_res_sd()<<setw(12)<<frecord[c].get_dx_res_sd();
	myfile<<setw(12)<<frecord[c].get_cdx_1()<<setw(12)<<frecord[c].get_cdy_1();
	myfile<<setw(12)<<frecord[c].get_cdx_x()<<setw(12)<<frecord[c].get_cdy_x();
	myfile<<setw(12)<<frecord[c].get_cdx_y()<<setw(12)<<frecord[c].get_cdy_y();
	myfile<<setw(12)<<frecord[c].get_cdx_x2()<<setw(12)<<frecord[c].get_cdy_x2();
	myfile<<setw(12)<<frecord[c].get_cdx_y2()<<setw(12)<<frecord[c].get_cdy_y2()<<"\n";
	cout<<".";
	}//for (int c=0; c<n_files;c++)
myfile.close();
cout<<"Saved.\n";
}//void FileTable::save(string file)

//main body
int main(int argc, char *argv[]) {

//setting default parameters
string time_table_name=def_time_table_name;
string images_table_name = def_images_table_name; 
float max_fwhm=def_max_fwhm;
int min_distance=def_min_distance;
int target_stars_count=def_target_stars_count;
int target_triangles_count=def_target_triangles_count;
float max_triangle_diff=def_max_triangle_diff;
int min_fits=def_min_fits;
float max_res_sd=def_max_res_sd;

//read parameters from command line switches
int first_LF_pos = 1;// search for first light image
string option, parameter1, parameter2;//strings to be read from shell

string *LF_file; //calibrated lightframe file

for (int c=1;c<argc-2;c++) {
	option = argv[c];
	parameter1 = argv[c+1]; parameter2 = argv[c+2];
	if (option =="-d") {stringstream(parameter1)>>min_distance;first_LF_pos+=2;continue;}
	if (option =="-q") {stringstream(parameter1)>>max_fwhm;first_LF_pos+=2;continue;}
	if (option =="-sc") {stringstream(parameter1)>>target_stars_count;first_LF_pos+=2;continue;}
	if (option =="-tc") {stringstream(parameter1)>>target_triangles_count;first_LF_pos+=2;continue;}
	if (option =="-td") {stringstream(parameter1)>>max_triangle_diff;first_LF_pos+=2;continue;}
	if (option =="-it") {images_table_name=parameter1;first_LF_pos+=2;continue;}
	if (option =="-tt") {time_table_name=parameter1;first_LF_pos+=2;continue;}
	}//for (int c=1;c<argc-2;c++)	

//correction of improper parameters
if (max_fwhm<1.1) {max_fwhm=1.1;cout<<"Maximum FWHM factor too low, corrected to "<<max_fwhm<<"\n";}
if (min_distance<2) {min_distance=2;cout<<"min_distance too low, corrected to "<<def_min_distance<<"\n";}
if (target_stars_count<25) {target_stars_count=25;cout<<"Target stars count too low, corrected to "<<target_stars_count<<"\n";}

int max_triangles_count=target_stars_count*(target_stars_count-1)*(target_stars_count-2)/12;//half of all possible triangles
if (target_triangles_count>max_triangles_count) {target_triangles_count=max_triangles_count;cout<<"Target triangles count too high, corrected to "<<target_triangles_count<<"\n";}
if (max_triangle_diff<0) {max_triangle_diff=0;cout<<"Maximum difference of triangle sides too low, corrected to "<<max_triangle_diff<<"\n";}

int n_LF = argc-first_LF_pos; //number of lightframes

LF_file = new string[n_LF];

//assign filenames
for (int c=0;c<n_LF;c++) {
	LF_file[c] = argv[first_LF_pos+c];
	}//for (int c=0;c<n_LF;c++)

ofstream time_table;
time_t seconds0,seconds;//time of run
time_table.open(time_table_name.data());
seconds0 = time(NULL);seconds=seconds0;time_table<<"start: "<<(seconds-seconds0)<<"\n"<<flush;

//load frames into table
FileTable ftab(n_LF,min_distance,target_stars_count,target_triangles_count,max_triangle_diff);
for (int c=0;c<n_LF;c++) {
	ftab.set_frecord (c,LF_file[c]);
	}//for (int c=0;c<n_LF;c++)

//calculate FWHM from all stars, set frame with best FWHM as reference
//deselect frames with high FWHM and low number of stars 
ftab.calculate_fwhm();
ftab.find_ref();
cout<<"Reference image index: "<<ftab.get_ref_index()<<"\n";
ftab.deselect_high_fwhm(max_fwhm);
ftab.deselect_low_stars(target_stars_count);
ftab.save(images_table_name);
seconds = time(NULL);time_table<<"calculate FWHM: "<<(seconds-seconds0)<<"\n"<<flush;

//calculate transformation parameters for each frame
cout<<"Calculating image transformation parameters: \n";
ftab.calculate_transformations();
ftab.save(images_table_name);
ftab.deselect_nonfits(min_fits);
ftab.deselect_bad_fits(max_res_sd);
ftab.save(images_table_name);
seconds = time(NULL);time_table<<"calculate transformation parameters: "<<(seconds-seconds0)<<"\n"<<flush;

//apply transformation on frames
cout<<"Transforming images:\n";
ftab.transform_images();
seconds = time(NULL);time_table<<"transform images: "<<(seconds-seconds0)<<"\n"<<flush;

return 0;
}//int main(int argc, char *argv[]) 

