/***BSD LICENSE***

Copyright (c) 2009-2010, Zdenek Sikac
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
    * The name of the author may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

***END OF BSD LICENSE***

version 0.3

The program imreg registers images. It calculates translation of each image with respect to reference image, transforms images and saves them.

USE:

imreg <switches> <input_file_1> <input_file_2> ... <input_file_N>

<file_1> - <file_N> images to be registered, <file_1> is used as reference.

Switches:

align parameters:
-box top_left_x top_left_y width height		define alignment box expressed as ratio of image dimensions (range 0 - 1). Default 0.45 0.45 0.1 0.1.
-m max_move					define maximum move of registered image (in pixels). Default 20.
-q						define minimum quality of registered image. Default 0.95. No effect in this version.

output tables:
-it	name of output file with table of images (default "images_table")
-tt	name of output file with time table (default no output)

*/


#include <iostream>
#include <fstream>
#include <Magick++.h>
#include <cmath>
#include <string>
#include <iomanip>
#include <algorithm>
#include <sstream>
#include "math.h"
#include <vector>

using namespace std;
using namespace Magick;

//default filenames parameters
const string def_time_table_name = "/dev/null";
const string def_images_table_name = "imreg_images_log"; 

//default alignment box size
const float def_x_pos = 0.45;
const float def_y_pos = 0.45;
const float def_box_width = 0.1;
const float def_box_height = 0.1;
const int def_max_move = 20;

//default quality filter
const float def_min_quality = 0.95;

//compare image to reference image and calculate offset and rotation
class FitImages {
Image *ref,*com;
int cdx_1,cdy_1;
float min_sum_diff;
public:
FitImages (Image&,Image&,float,float,float,float,int,int,int);
int get_cdx_1() {return(cdx_1);};
int get_cdy_1() {return(cdy_1);};
float get_min_sum_diff() {return(min_sum_diff);};
};

FitImages::FitImages (Image& ref,Image& com,float rel_x_pos,float rel_y_pos,float rel_box_width,float rel_box_height,int max_move,int dx_last,int dy_last) {

int w = ref.columns(); int h = ref.rows();
int x_pos = w*rel_x_pos; int y_pos = h*rel_y_pos; int box_width = w*rel_box_width;int box_height = h*rel_box_height;
int box_area = box_width*box_height;
int min_sum_diff=MaxRGB;

ref.modifyImage();
PixelPacket *ref_cache=ref.getPixels(x_pos,y_pos,box_width,box_height);//pixel cache from image ref

for (int dx=-max_move;dx<=max_move;dx++) {
	for (int dy=-max_move;dy<=max_move;dy++) {
		com.modifyImage();
		PixelPacket *com_cache=com.getPixels(x_pos+dx+dx_last,y_pos+dy+dy_last,box_width,box_height);//pixel cache from image com
		float sum_diff=0;		
		for (int c=0;c<box_area;c++) {
			sum_diff = sum_diff+abs(ref_cache[c].green-com_cache[c].green);
			}//for (int c=0;c<box_area;c++)
		sum_diff=sum_diff/box_area;
		if (sum_diff<min_sum_diff) {min_sum_diff = sum_diff;cdx_1 = dx+dx_last;cdy_1 = dy+dy_last;}
		}//for (int dy=-max_move;dy<=max_move;dy++)
	}//for (int dx=-max_move;dx<=max_move;dx++)
}//class FitImages

//class for storing of image data
class FileRecord {
string filename;
bool use, ref;
float cdx_1,cdy_1;

public:
void initialize (string);
string get_filename() {return(filename);}
float get_cdx_1() {return(cdx_1);}
float get_cdy_1() {return(cdy_1);}
bool get_use() {return(use);}
bool get_ref() {return(ref);}
void set_ref(bool new_ref) {ref = new_ref;}
void set_use(bool new_use) {use = new_use;}
void set_trafo_parameters(float,float);
void transform_image();
FileRecord();
};//class FileRecord

FileRecord::FileRecord() {
filename="";use=true;ref=false;
cdx_1=0;cdy_1=0;
}//FileRecord::FileRecord()

void FileRecord::initialize(string filename_i) {
filename=filename_i;
use=true;ref=false;
cdx_1=0;cdy_1=0;
}//void FileRecord::intialize(string filename_i)

void FileRecord::set_trafo_parameters (float cdx_1_i,float cdy_1_i) {

cdx_1=cdx_1_i;cdy_1=cdy_1_i;
}//void FileRecord::set_trafo_parameters(float,float)

void FileRecord::transform_image() {
Image original_image (filename);
int w = original_image.columns();int h = original_image.rows();
int area=w*h;

Image transformed_image(Geometry(w,h),Color(0,0,0,MaxRGB));

float x_tmp,y_tmp;

original_image.modifyImage();
Pixels original_cache(original_image);
PixelPacket *original_pixels=original_cache.get(0,0,w,h);//pixel cache from original image

transformed_image.modifyImage();
Pixels transformed_cache(transformed_image);
PixelPacket *transformed_pixels=transformed_cache.get(0,0,w,h);//pixel cache from image

int cx,cy;
int pos_tmp;

cout<<"Transforming image "<<filename;

for (int c=0;c<area;c++) {
	counter (c,area,10);	
	cx=c%w;	
	cy=c/w;
	x_tmp = cx+cdx_1;
	y_tmp = cy+cdy_1;
	pos_tmp = round(x_tmp)+round(y_tmp)*w;
	if ((pos_tmp<0)||(pos_tmp>area)) continue;
	transformed_pixels[c] = original_pixels[pos_tmp];
	}//for (int c=0;c<area;c++)

original_cache.sync();
transformed_cache.sync();

transformed_image.matte(false);
transformed_image.write("reg_"+filename);
cout<<"Saved as reg_"<<filename<<".\n";
}//void FileRecord::transform_image()

//table for storing of FileRecords
class FileTable {
int n_files;FileRecord *frecord;float x_pos;float y_pos;float box_width;float box_height;int max_move;
int ref_index;

public:
int get_n_files() {return (n_files);};
void calculate_transformations();
FileTable(int,float,float,float,float,int);
~FileTable();
void set_frecord(int, string);
void save(string);
int get_ref_index() {return(ref_index);}
FileRecord get_frecord(int c) {return frecord[c];}
void transform_images();
};//class FileTable;

FileTable::FileTable (int n_files_i,float x_pos_i,float y_pos_i,float box_width_i,float box_height_i,int max_move_i) {
n_files = n_files_i;ref_index = 0;
frecord = new FileRecord [n_files];
x_pos = x_pos_i;
y_pos = y_pos_i;
box_width = box_width_i;
box_height = box_height_i;
max_move = max_move_i;
}//FileTable::FileTable (int,float,float,float,float,int)

FileTable::~FileTable() {
delete [] frecord;
}//FileTable::~FileTable()

//assign filenames to file records
void FileTable::set_frecord (int n, string filename) {
frecord[n].initialize (filename);
}//void FileTable::set_frecord (int n, string filename)

//calculate transformation parameters from reference image
void FileTable::calculate_transformations() {
string ref_file, com_file;
ref_file = frecord[ref_index].get_filename();
cout<<"Using image "<<ref_file<<" as reference.\n";
Image ref_image(ref_file);
int dx_last=0,dy_last=0;
 
for (int c=0; c<n_files;c++) {
	if(c==ref_index) continue; //skip calculation if current image is reference
	if(frecord[c].get_use()==false) continue; //skip calculation if current image is marked as unused
	com_file=frecord[c].get_filename();
	cout<<"Calculating offsets of image "<<com_file<<" from reference image.\n";
	Image com_image(com_file);

	FitImages fit(ref_image,com_image,x_pos,y_pos,box_width,box_height,max_move,dx_last,dy_last);
	frecord[c].set_trafo_parameters	(fit.get_cdx_1(),fit.get_cdy_1());
	dx_last=fit.get_cdx_1();dy_last=fit.get_cdy_1();
	}//for (int c=0; c<n_files;c++)
}//void FileTable::calculate_transformations()

//register images
void FileTable::transform_images() {
for (int c=0; c<n_files;c++) {
	if(frecord[c].get_use()==false) continue;
	frecord[c].transform_image();
	}//for (int c=0; c<n_files;c++)
}//void FileTable::transform_images()

//save file table into file
void FileTable::save(string file) {
ofstream myfile;
const char *cfilename = file.c_str();

myfile.open (cfilename);

cout<<"Saving table of images to file "<<file;
myfile<<setw(5)<<"N:"<<setw(20)<<"File:"<<setw(8)<<"Ref:"<<setw(8)<<"Used:"<<setw(8);
myfile<<setw(12)<<"cdx_1:"<<setw(12)<<"cdy_1:\n";

for (int c=0; c<n_files;c++) {
	myfile<<setw(5)<<fixed<<setprecision(0)<<noshowpos<<c;
	myfile<<setw(20)<<frecord[c].get_filename();
	myfile<<setw(8)<<setprecision(0)<<frecord[c].get_ref()<<setw(8)<<frecord[c].get_use();
	myfile<<setw(12)<<frecord[c].get_cdx_1()<<setw(12)<<frecord[c].get_cdy_1()<<"\n";
	cout<<".";
	}//for (int c=0; c<n_files;c++)
myfile.close();
cout<<"Saved.\n";
}//void FileTable::save(string file)

//main body
int main(int argc, char *argv[]) {

//setting default parameters
string time_table_name=def_time_table_name;
string images_table_name = def_images_table_name; 
float x_pos=def_x_pos;
float y_pos=def_y_pos;
float box_width=def_box_width;
float box_height=def_box_height;
int max_move=def_max_move;
float min_quality=def_min_quality;

//read parameters from command line switches
int first_LF_pos = 1;// search for first light image
string option, parameter;//strings to be read from shell

string *LF_file; //calibrated lightframe file

for (int c=1;c<argc-1;c++) {
	option = argv[c];
	parameter = argv[c+1];
	if (option =="-box") {
		stringstream(argv[c+1])>>x_pos;
		stringstream(argv[c+2])>>y_pos;
		stringstream(argv[c+3])>>box_width;
		stringstream(argv[c+4])>>box_height;
		first_LF_pos+=5;
		continue;
		}
	if (option =="-m") {stringstream(parameter)>>max_move;first_LF_pos+=2;continue;}
	if (option =="-q") {stringstream(parameter)>>min_quality;first_LF_pos+=2;continue;}
	if (option =="-it") {images_table_name=parameter;first_LF_pos+=2;continue;}
	if (option =="-tt") {time_table_name=parameter;first_LF_pos+=2;continue;}
	}	

int n_LF = argc-first_LF_pos; //number of lightframes

LF_file = new string[n_LF];

//assign filenames
for (int c=0;c<n_LF;c++) {
	LF_file[c] = argv[first_LF_pos+c];
	}//for (int c=0;c<n_LF;c++)

ofstream time_table;
time_t seconds0,seconds;//time of run
time_table.open(time_table_name.data());
seconds0 = time(NULL);seconds=seconds0;time_table<<"start: "<<(seconds-seconds0)<<"\n"<<flush;

//load frames into table
FileTable ftab(n_LF,x_pos,y_pos,box_width,box_height,max_move);
for (int c=0;c<n_LF;c++) {
	ftab.set_frecord (c,LF_file[c]);
	}//for (int c=0;c<n_LF;c++)

//calculate transformation parameters for each frame
cout<<"Calculating image transformation parameters: \n";
ftab.calculate_transformations();
ftab.save(images_table_name);
seconds = time(NULL);time_table<<"calculate transformation parameters: "<<(seconds-seconds0)<<"\n"<<flush;

//apply transformation on frames
cout<<"Transforming images:\n";
ftab.transform_images();
seconds = time(NULL);time_table<<"transform images: "<<(seconds-seconds0)<<"\n"<<flush;

return 0;
}//int main(int argc, char *argv[]) 

