/***BSD LICENSE***

Copyright (c) 2010, Zdenek Sikac
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
    * The name of the author may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

***END OF BSD LICENSE***


version 0.1

The program imstats creates table with statistics for each image.

USE:

imstats <switches> <input_file_1> <input_file_2> ... <input_file_N>

<input_file_1> - <input_file_N> images to be registered, image with best quality (lowest FWHM) is used as reference.

Switches:

output log tables:
-it [string]	name of result table of images (default string = "imstats_results")
-tt [string]	name of output file with time table (default no output)

*/

#include <iostream>
#include <fstream>
#include <Magick++.h>
#include <cmath>
#include <string>
#include <iomanip>
#include <algorithm>
#include <sstream>
#include "math.h"
#include <vector>

using namespace std;
using namespace Magick;

//default filenames parameters
const string def_time_table_name = "/dev/null";
const string def_images_table_name = "imstats_results"; 

class ImageAnalysis {
void initialize();
int width,height,area,depth,nil,IA_median;
int *green;
float IA_stdev,IA_average;
Image image;

public:
ImageAnalysis (string);
~ImageAnalysis();
float get_average() {return(IA_average);}
float get_stdev() {return(IA_stdev);}
int get_median() {return(IA_median);}
};

ImageAnalysis::~ImageAnalysis () {
delete [] green;
}//ImageAnalysis::~ImageAnalysis ()

ImageAnalysis::ImageAnalysis (string filename) {
cout<<"Opening file "<<filename<<"\n";
Image new_image;
try {
	new_image.read(filename);
	}

catch ( Exception &error_ ) {
	cout << "Caught exception: " << error_.what() << endl;
	
	}//try
image = new_image;
initialize();
}

//calculate parameters of image
void ImageAnalysis::initialize () {
width=image.columns();height=image.rows();
area=width*height; //area = number of pixels
const float base = 2;

depth=image.depth();nil=int(pow(base,depth));//nil=number of intensity levels

image.modifyImage();
Pixels pixelcache(image);
PixelPacket *pixels=pixelcache.get(0,0,width,height);//pixel cache from image

//get array of pixel intensities in sample from image
green = new int [area];
for (int c=0;c<area;c++) {
	green[c]=pixels[c].green;
	};//for (int c=0;c<area;c++)

pixelcache.sync();

IA_average = average (green, area);
IA_stdev = stdev (green, area);
IA_median = quantil (green, area, 0.5);
pixelcache.sync();
}//ImageAnalysis::ImageAnalysis (string filename)

//class for storing of image data
class FileRecord {
string FR_filename;
int FR_median;
float FR_average,FR_stdev;

public:
void initialize (string);
string get_filename() {return(FR_filename);}
int get_median() {return(FR_median);}
float get_average() {return(FR_average);}
float get_stdev() {return(FR_stdev);}
FileRecord();
};//class FileRecord

FileRecord::FileRecord() {
FR_filename="";FR_median=0;FR_average=0;FR_stdev=0;
}//FileRecord::FileRecord()

void FileRecord::initialize(string filename) {
FR_filename=filename;
ImageAnalysis ia(filename);
FR_median=ia.get_median();
FR_average=ia.get_average();
FR_stdev=ia.get_stdev();
}//void FileRecord::intialize(string filename)

//table for storing of FileRecords
class FileTable {
int FT_n_files;FileRecord *frecord;

public:
int n_files() {return (FT_n_files);};
void analyze_images();
FileTable(int);
~FileTable();
void set_frecord(int, string);
void save(string);
};//class FileTable;

FileTable::FileTable (int n_files) {
FT_n_files = n_files;
frecord = new FileRecord [FT_n_files];
}//FileTable::FileTable (int n_files)

FileTable::~FileTable() {
delete [] frecord;
}//FileTable::~FileTable()

//assign filenames to file records
void FileTable::set_frecord (int n, string filename) {
frecord[n].initialize (filename);
}//void FileTable::set_frecord (int n, string filename)

void FileTable::save(string file) {
ofstream myfile;
const char *cfilename = file.c_str();

myfile.open (cfilename);

cout<<"Saving table of images to file "<<file;
myfile<<setw(5)<<"N:"<<setw(20)<<"File:";
myfile<<setw(12)<<"Average:"<<setw(12)<<"SD:"<<setw(12)<<"Median:\n";

for (int c=0; c<FT_n_files;c++) {
	myfile<<setw(5)<<fixed<<setprecision(0)<<noshowpos<<c;
	myfile<<setw(20)<<frecord[c].get_filename();
	myfile<<setw(12)<<setprecision(2)<<frecord[c].get_average()<<setw(12)<<frecord[c].get_stdev();
	myfile<<setw(12)<<setprecision(1)<<frecord[c].get_median()<<"\n";
	cout<<".";
	}//for (int c=0; c<n_files;c++)
myfile.close();
cout<<"Saved.\n";
}//void FileTable::save(string file)

//main body
int main(int argc, char *argv[]) {

//setting default parameters
string time_table_name=def_time_table_name;
string images_table_name = def_images_table_name; 

//read parameters from command line switches
int first_IF_pos = 1;// search for first light image
string option, parameter1, parameter2;//strings to be read from shell

string *IF_file; //calibrated lightframe file

for (int c=1;c<argc-2;c++) {
	option = argv[c];
	parameter1 = argv[c+1]; parameter2 = argv[c+2];
	if (option =="-it") {images_table_name=parameter1;first_IF_pos+=2;continue;}
	if (option =="-tt") {time_table_name=parameter1;first_IF_pos+=2;continue;}
	}//for (int c=1;c<argc-2;c++)	

int n_IF = argc-first_IF_pos; //number of lightframes

IF_file = new string[n_IF];

//assign filenames
for (int c=0;c<n_IF;c++) {
	IF_file[c] = argv[first_IF_pos+c];
	}//for (int c=0;c<n_LF;c++)

ofstream time_table;
time_t seconds0,seconds;//time of run
time_table.open(time_table_name.data());
seconds0 = time(NULL);seconds=seconds0;time_table<<"start: "<<(seconds-seconds0)<<"\n"<<flush;

//load frames into table
FileTable ftab(n_IF);
for (int c=0;c<n_IF;c++) {
	ftab.set_frecord (c,IF_file[c]);
	}//for (int c=0;c<n_LF;c++)

ftab.save(images_table_name);
seconds = time(NULL);time_table<<"end: "<<(seconds-seconds0)<<"\n"<<flush;
return 0;
}//int main(int argc, char *argv[]) 

