/***BSD LICENSE***

Copyright (c) 2009-2010, Zdenek Sikac
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
    * The name of the author may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

***END OF BSD LICENSE***


version 0.6

stack produces stacked image from several images.

USE:

stack <SWITCHES> <input_file_1> <input_file_2> ... <input_file_N> <output_file>

Input images must have equal size. The images are processed in tiles so any number of images may be stacked without huge requirements for RAM. Different stacking algorithms can be used.


SWITCHES:

algorithm:
-al 	stacking algorithm (0 = average, 1 = median, 2 = sigma-clipping, 3 = maximum) (default 0)
-slo	low limit for sigma-clipping (default 2)
-shi	high limit for sigma-clipping (default 2)

output log:
-tt [string]	name of output file with time table (default no output)

NOTE:
The images should be registered before stacking, e.g. by dsireg or imreg.

*/

#include <iostream>
#include <fstream>
#include <Magick++.h>
#include <cmath>
#include <string>
#include <iomanip>
#include <sstream>
#include "math.h"

using namespace std;
using namespace Magick;

int main(int argc, char *argv[]) {

string *image_file;
string output_image_file;
Image my_image;
int red_out = 0,green_out = 0,blue_out = 0;
int px,py,nx,ny,parea;
int px_cur,py_cur;
int *red;int *green;int *blue;

//defaults
int algorithm=0; 
float slo=2,shi=2;

string time_table_name = "/dev/null";

unsigned short *red_array;
unsigned short *green_array;
unsigned short *blue_array;

PixelPacket *my_cache;

//read parameters from command line switches
int first_image_pos = 1;
string option, parameter;

for (int c=1;c<argc-1;c++) {
	option = argv[c];
	parameter = argv[c+1];
	if (option =="-al") {stringstream(parameter)>>algorithm;first_image_pos+=2;continue;}
	if (option =="-slo") {stringstream(parameter)>>slo;first_image_pos+=2;continue;}
	if (option =="-shi") {stringstream(parameter)>>shi;first_image_pos+=2;continue;}
	if (option =="-tt") {time_table_name=parameter;first_image_pos+=2;continue;}
	}	

int n = argc-first_image_pos-1;

cout<<"Number of images = "<<n<<"\n";

image_file = new string[n];

//assign filenames
for (int c=0;c<n;c++) {
	image_file[c] = argv[first_image_pos+c];
	}

output_image_file = argv[argc-1];

//time log
ofstream time_table;
time_t seconds0,seconds;//time of run
time_table.open(time_table_name.data());
seconds0 = time(NULL);seconds=seconds0;time_table<<"start: "<<(seconds-seconds0)<<"\n"<<flush;

//get geometry of images and define size of processing area
try {
	my_image.read(image_file[0]);
	}

catch ( Exception &error_ ) {
cout << "Caught exception: " << error_.what() << endl;
	}

int const width=my_image.columns();int const height=my_image.rows();
px = py = 4096 / sqrt(n);
nx = width/px;
ny = height/py;

cout<<"Processing area size: x = "<<px<<", y = "<<py<<"\n";

//allocate space for colors of a processed pixel
red = new int [n];
green = new int [n];
blue = new int [n];

//create target averaged image
Image image_out (Geometry(width,height),Color(0,0,0));
image_out.type(TrueColorType);
PixelPacket *cache_out;

image_out.modifyImage();

for (int cx=0;cx<=nx;cx++) {
	
	px_cur = px;
	if ((cx*px+px_cur)>width) px_cur = width-cx*px;
	
	for (int cy=0;cy<=ny;cy++) {
		
		py_cur = py;
		if ((cy*py+py_cur)>height) py_cur = height-cy*py;
		
		parea = px_cur*py_cur;
		cout<<"Processing area: x = "<<cx*px<<" to "<<cx*px+px_cur<<", y = ";
		cout<<cy*py<<" to "<<cy*py+py_cur<<"\n";
		red_array = new unsigned short [parea*n]; 
		green_array = new unsigned short [parea*n]; 
		blue_array = new unsigned short [parea*n]; 
		cache_out = image_out.getPixels(cx*px,cy*py,px_cur,py_cur); //cache of new averaged image
		
		for (int c=0;c<n;c++) {
			try {
				my_image.read(image_file[c]);
				}
			catch ( Exception &error_ ) {
			cout << "Caught exception: " << error_.what() << endl;
				}
			my_cache=my_image.getPixels(cx*px,cy*py,px_cur,py_cur);
			
			for (int p=0;p<parea;p++) {
				red_array[c*parea+p] = (my_cache[p].red);
				green_array[c*parea+p] = (my_cache[p].green);
				blue_array[c*parea+p] = (my_cache[p].blue);
				}
			}
//color values are copied from processed area to temporary arrays
		for (int p=0;p<parea;p++) {
			for (int c=0;c<n;c++) {
				red [c] = red_array[c*parea+p];
				green [c] = green_array[c*parea+p];
				blue [c] = blue_array[c*parea+p];
				}

			//average values of color intensities are calculated for each pixel	
			if (algorithm==0) {
				red_out = average (red, n);
				green_out = average (green, n);
				blue_out = average (blue, n);
				}

			//median values of color intensities are calculated for each pixel	
			if (algorithm==1) {
				red_out = median (red, n);
				green_out = median (green, n);
				blue_out = median (blue, n);
				}

			//sigma-clipped average values of color intensities are calculated for each pixel	
			if (algorithm==2) {
				red_out = sigma_clip (red, n, slo, shi);
				green_out = sigma_clip (green, n, slo, shi);
				blue_out = sigma_clip (blue, n, slo, shi);
				}

			//sigma-clipped average values of color intensities are calculated for each pixel	
			if (algorithm==3) {
				red_out = max (red, n);
				green_out = max (green, n);
				blue_out = max (blue, n);
				}

			cache_out[p]=Color(red_out,green_out,blue_out);
			}
//temporary arrays are deallocated
		delete[] red_array;		
		delete[] green_array;		
		delete[] blue_array;
//pixel cache is synchronized with output image
		image_out.syncPixels();		
		}
	}

delete [] red;
delete [] green;
delete [] blue;

//saving averaged image into file
cout<<"Saving image to file "<<output_image_file<<".\n";
image_out.write(output_image_file);

seconds = time(NULL);time_table<<"end: "<<(seconds-seconds0)<<"\n"<<flush;

return 0;
}


